<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:file="http://expath.org/ns/file"
                xmlns:map="http://www.w3.org/2005/xpath-functions/map"
                xmlns:saxon="http://saxon.sf.net/"
                xmlns:shared="urn:shared-functions"
                xmlns:thread="java:java.lang.Thread"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                version="2.0">
   <xsl:output method="xml" version="1.0" encoding="UTF-8" indent="yes" saxon:indent-spaces="0" saxon:line-length="100000" saxon:attribute-order="*"/>
   <xsl:variable name="pipe" select="'|'"/>
   <xsl:variable name="separator" select="'-'"/>
   <xsl:variable name="EOL" select="'&#10;'"/>
   <xsl:variable name="TAB" select="'&#x9;'"/>
   <xsl:key name="with-key" match="*" use="@key"/>

   <xsl:function name="shared:always-false" as="xs:boolean">
      <xsl:sequence select="current-date() lt xs:date('2000-01-01')"/>
   </xsl:function>

   <xsl:function name="shared:chars" as="xs:string*">
      <!-- Same as functx:chars -->
      <xsl:param name="arg" as="xs:string?"/>
      <xsl:sequence select="for $ch in string-to-codepoints($arg) return codepoints-to-string($ch)"/>
   </xsl:function>

   <xsl:function name="shared:create-dir">
      <xsl:param name="dir" as="xs:string"/>
      <xsl:sequence select="shared:log('DEBUG', concat('Creating ', $dir))"/>
      <xsl:sequence select="file:create-dir($dir)"/>
   </xsl:function>

   <xsl:function name="shared:delete">
      <xsl:param name="path" as="xs:string"/>
      <xsl:sequence select="shared:delete($path, true())"/>
   </xsl:function>

   <xsl:function name="shared:delete">
      <xsl:param name="path" as="xs:string"/>
      <xsl:param name="recursive" as="xs:boolean"/>
      <xsl:if test="file:exists($path)">
         <xsl:sequence select="shared:log('DEBUG', concat('Deleting ', $path))"/>
         <xsl:sequence select="file:delete($path, $recursive)"/>
      </xsl:if>
   </xsl:function>

   <xsl:function name="shared:doc" as="document-node()?">
      <xsl:param name="uri" as="xs:string"/>
      <xsl:choose>
         <xsl:when test="doc-available($uri)">
            <xsl:sequence select="saxon:discard-document(doc($uri))"/>
         </xsl:when>
         <xsl:otherwise>
            <xsl:sequence select="shared:log('DEBUG', concat('*** MISSING ', $uri))"/>
         </xsl:otherwise>
      </xsl:choose>
   </xsl:function>

   <xsl:function name="shared:extract-validities" as="element()*">
      <xsl:param name="entityVersions" as="element()*"/>
      <xsl:param name="refValidFromDate" as="xs:string"/>
      <xsl:param name="refValidToDate" as="xs:string"/>
      <xsl:variable name="validFromDates" as="xs:string*" select="distinct-values($entityVersions/@validFromDate[. gt $refValidFromDate and . le $refValidToDate])"/>
      <xsl:variable name="validToDates" as="xs:string*" select="distinct-values($entityVersions/@validToDate[. ge $refValidFromDate and . lt $refValidToDate])"/>
      <xsl:variable name="shiftedValidFromDates" as="xs:string*" select="$validFromDates!shared:previous-day(.)"/>
      <xsl:variable name="shiftedValidToDates" as="xs:string*" select="$validToDates!shared:next-day(.)"/>
      <xsl:variable name="mergedValidFromDates" as="xs:string*" select="distinct-values(($refValidFromDate, $validFromDates, $shiftedValidToDates))"/>
      <xsl:variable name="mergedValidToDates" as="xs:string*" select="distinct-values(($refValidToDate, $validToDates, $shiftedValidFromDates))"/>
      <xsl:variable name="dates" as="xs:string*" select="sort(($mergedValidFromDates, $mergedValidToDates))"/>
      <xsl:for-each select="$dates">
         <xsl:variable name="index" select="position()"/>
         <xsl:if test="($index mod 2) = 1">
            <xsl:sequence select="shared:validity(., $dates[$index + 1])"/>
         </xsl:if>
      </xsl:for-each>
   </xsl:function>

   <xsl:function name="shared:file-extension" as="xs:string">
      <xsl:param name="uri" as="xs:string"/>
      <!-- \ escape char in regex is needed for '.' -->
      <xsl:sequence select="tokenize($uri, '\.')[last()]"/>
   </xsl:function>

   <xsl:function name="shared:get-validities" as="element()*">
      <xsl:param name="element" as="element()"/>
      <xsl:param name="entityVersions" as="element()*"/>
      <xsl:for-each select="$element">
         <xsl:sequence select="shared:extract-validities($entityVersions, @validFromDate, @validToDate)"/>
      </xsl:for-each>
   </xsl:function>

   <xsl:function name="shared:get-validities" as="element()*">
      <xsl:param name="element" as="element()"/>
      <xsl:param name="entityVersions" as="element()*"/>
      <xsl:param name="purgeDate" as="xs:string"/>
      <xsl:sequence select="shared:get-validities($element, $entityVersions)[@validToDate ge $purgeDate]"/>
   </xsl:function>

   <xsl:function name="shared:is-absolute-uri" as="xs:boolean">
      <xsl:param name="uri" as="xs:string?"/>
      <xsl:sequence select="matches($uri, '^[a-z]+:')"/>
   </xsl:function>

   <xsl:function name="shared:list" as="xs:string*">
      <xsl:param name="dir" as="xs:string"/>
      <xsl:param name="recursive" as="xs:boolean"/>
      <xsl:param name="pattern" as="xs:string"/>
      <xsl:if test="file:exists($dir)">
         <xsl:for-each select="file:list($dir, $recursive, $pattern)">
            <xsl:sort select="."/>
            <xsl:sequence select="translate(., '\', '/')"/>
         </xsl:for-each>
      </xsl:if>
   </xsl:function>

   <xsl:function name="shared:list-files" as="xs:string*">
      <xsl:param name="dir" as="xs:string"/>
      <xsl:param name="recursive" as="xs:boolean"/>
      <xsl:param name="pattern" as="xs:string"/>
      <xsl:sequence select="shared:list($dir, $recursive, $pattern)[file:is-file(concat($dir, .))]"/>
   </xsl:function>

   <xsl:function name="shared:log">
      <xsl:param name="level" as="xs:string"/>
      <xsl:param name="message" as="item()*"/>
      <xsl:variable name="timestamp" select="string(saxon:timestamp())"/>
      <xsl:message select="string-join((substring($timestamp, 1, 10), substring($timestamp, 12, 8), concat('[', $level, ']'), serialize($message)), ' ')"/>
   </xsl:function>

   <xsl:function name="shared:log-error">
      <xsl:param name="description" as="xs:string"/>
      <xsl:param name="module" as="xs:string"/>
      <xsl:param name="line-number" as="xs:integer?"/>
      <xsl:sequence select="shared:log('ERROR', concat('description: ', $description))"/>
      <xsl:sequence select="shared:log('ERROR', concat('module: ', $module))"/>
      <xsl:sequence select="shared:log('ERROR', concat('line-number: ', $line-number))"/>
   </xsl:function>

   <xsl:function name="shared:log-execution-time">
      <xsl:param name="startTime" as="xs:dateTime"/>
      <!-- Use xs:dayTimeDuration instead of xs:duration as said in Saxon's documentation -->
      <xsl:sequence select="shared:log('DEBUG', concat('Execution time: ', (saxon:timestamp() - $startTime) div xs:dayTimeDuration('PT1S'), 's'))"/>
   </xsl:function>

   <xsl:template name="shared:log-stylesheet-params">
      <xsl:param name="stylesheet-params" as="map(*)?"/>
      <xsl:sequence select="shared:log-stylesheet-params($stylesheet-params)"/>
   </xsl:template>

   <xsl:function name="shared:log-stylesheet-params">
      <xsl:param name="stylesheet-params" as="map(*)?"/>
      <xsl:if test="exists($stylesheet-params)">
         <xsl:for-each select="map:keys($stylesheet-params)">
            <xsl:sort select="string(.)"/>
            <xsl:variable name="value" select="map:get($stylesheet-params, .)"/>
            <xsl:variable name="parameter-value"
                          select="
                          if (empty($value)) then '()' else
                          if ($value instance of map(*)+) then concat('map(#', count($value), ')') else
                          if ($value instance of node()+) then string-join((for $v in $value return ($v/saxon:system-id(), path($v))[1]), ' ') else
                          if (contains(lower-case(string(.)), 'password')) then replace($value, '.', '*') else
                          string-join($value, ' ')"/>
            <xsl:sequence select="shared:log('INFO', concat($TAB, 'parameter ',., '=', $parameter-value))"/>
         </xsl:for-each>
      </xsl:if>
   </xsl:function>

   <xsl:function name="shared:log-transform-options">
      <xsl:param name="options" as="map(*)"/>
      <xsl:variable name="source-location" select="map:get($options, 'source-location')"/>
      <xsl:variable name="base-output-uri" select="map:get($options, 'base-output-uri')"/>
      <xsl:variable name="stylesheet-location" select="map:get($options, 'stylesheet-location')"/>
      <xsl:variable name="stylesheet-params" select="map:get($options, 'stylesheet-params')"/>
      <xsl:choose>
         <xsl:when test="exists($source-location) and exists($base-output-uri)">
            <xsl:sequence select="shared:log('INFO', concat('Transforming ',$source-location))"/>
            <xsl:sequence select="shared:log('INFO', concat($TAB, 'into ',$base-output-uri))"/>
            <xsl:sequence select="shared:log('INFO', concat($TAB, 'using ',$stylesheet-location))"/>
         </xsl:when>
         <xsl:when test="exists($source-location) and empty($base-output-uri)">
            <xsl:sequence select="shared:log('INFO', concat('Transforming ',$source-location))"/>
            <xsl:sequence select="shared:log('INFO', concat($TAB, 'using ',$stylesheet-location))"/>
         </xsl:when>
         <xsl:when test="empty($source-location) and exists($base-output-uri)">
            <xsl:sequence select="shared:log('INFO', concat('Generating ',$base-output-uri))"/>
            <xsl:sequence select="shared:log('INFO', concat($TAB, 'using ',$stylesheet-location))"/>
         </xsl:when>
         <xsl:otherwise>
            <xsl:sequence select="shared:log('INFO', concat('Launching ',$stylesheet-location))"/>
         </xsl:otherwise>
      </xsl:choose>
      <xsl:sequence select="shared:log-stylesheet-params($stylesheet-params)"/>
   </xsl:function>

   <xsl:function name="shared:next-day" as="xs:string">
      <xsl:param name="date"/>
      <xsl:sequence select="shared:shift-day($date, 1)"/>
   </xsl:function>

   <xsl:function name="shared:open-file" as="node()?">
      <xsl:param name="pathAndFileName" as="xs:string?"/>
      <xsl:for-each select="$pathAndFileName">
         <xsl:variable name="uri" select="shared:path-to-uri(.)"/>
         <xsl:sequence select="shared:doc($uri)"/>
      </xsl:for-each>
   </xsl:function>

   <xsl:function name="shared:open-file" as="document-node()?">
      <xsl:param name="baseURI" as="xs:string?"/>
      <xsl:param name="fileName" as="xs:string?"/>
      <xsl:if test="exists($baseURI) and exists($fileName)">
         <xsl:variable name="pathAndFileName" select="concat($baseURI, $fileName)"/>
         <xsl:sequence select="shared:open-file($pathAndFileName)"/>
      </xsl:if>
   </xsl:function>

   <xsl:function name="shared:open-files" as="document-node()*">
      <xsl:param name="baseURI" as="xs:string"/>
      <xsl:param name="pattern" as="xs:string"/>
      <xsl:variable name="uri" select="shared:path-to-uri($baseURI)"/>
      <xsl:variable name="files" select="collection(concat($uri, '?select=', $pattern))"/>
      <xsl:sequence select="shared:log('DEBUG', concat('Nb of loaded files = ', count($files)))"/>
      <xsl:for-each select="$files">
         <xsl:sort select="saxon:system-id()"/>
         <xsl:sequence select="saxon:discard-document(.)"/>
      </xsl:for-each>
   </xsl:function>

   <xsl:function name="shared:path-to-uri" as="xs:string">
      <xsl:param name="path" as="xs:string"/>
      <xsl:sequence select="file:path-to-uri($path)"/>
   </xsl:function>

   <xsl:function name="shared:path-to-uri" as="xs:string">
      <xsl:param name="baseURI" as="xs:string"/>
      <xsl:param name="fileName" as="xs:string"/>
      <xsl:variable name="path" select="concat($baseURI, $fileName)"/>
      <xsl:sequence select="file:path-to-uri($path)"/>
   </xsl:function>

   <xsl:function name="shared:previous-day" as="xs:string">
      <xsl:param name="date"/>
      <xsl:sequence select="shared:shift-day($date, -1)"/>
   </xsl:function>

   <xsl:function name="shared:shift-day" as="xs:string">
      <xsl:param name="date"/>
      <xsl:param name="nbOfDays"/>
      <xsl:variable name="pDate" select="xs:date($date)"/>
      <xsl:variable name="pNbOfDays" select="xs:integer($nbOfDays)"/>
      <xsl:choose>
         <xsl:when test="$pNbOfDays gt 0">
            <!-- Add abs to avoid compilation warning -->
            <xsl:value-of select="$pDate + xs:dayTimeDuration(concat('P', abs($pNbOfDays), 'D'))"/>
         </xsl:when>
         <xsl:when test="$pNbOfDays lt 0">
            <xsl:value-of select="$pDate - xs:dayTimeDuration(concat('P', abs($pNbOfDays), 'D'))"/>
         </xsl:when>
         <xsl:otherwise>
            <xsl:value-of select="$pDate"/>
         </xsl:otherwise>
      </xsl:choose>
   </xsl:function>

   <xsl:function name="shared:sleep">
      <xsl:param name="seconds" as="xs:integer"/>
      <xsl:sequence select="shared:log('DEBUG', concat('Sleeping for ', $seconds, ' seconds...'))"/>
      <xsl:sequence select="thread:sleep(1000*$seconds)"/>
   </xsl:function>

   <xsl:function name="shared:to-folder" as="xs:string?">
      <xsl:param name="property" as="xs:string?"/>
      <xsl:for-each select="$property">
         <xsl:for-each select="translate(., '\', '/')">
            <xsl:sequence select="if (ends-with(., '/')) then . else concat(., '/')"/>
         </xsl:for-each>
      </xsl:for-each>
   </xsl:function>

   <xsl:template name="shared:transform">
      <xsl:param name="options" as="map(*)"/>
      <xsl:variable name="startTime" select="saxon:timestamp()"/>
      <xsl:sequence select="$startTime[shared:always-false()]"/>
      <xsl:variable name="extended-options" select="map:merge((map:entry('delivery-format', 'serialized'), $options))"/>
      <xsl:call-template name="shared:write">
         <xsl:with-param name="output" select="shared:transform($extended-options)"/>
      </xsl:call-template>
      <xsl:sequence select="shared:log-execution-time($startTime)"/>
   </xsl:template>

   <xsl:function name="shared:transform" as="map(*)">
      <xsl:param name="options" as="map(*)"/>
      <xsl:sequence select="shared:log-transform-options($options)"/>
      <xsl:sequence select="transform($options)"/>
   </xsl:function>

   <xsl:function name="shared:validity" as="element()">
      <xsl:param name="validFromDate" as="xs:string"/>
      <xsl:param name="validToDate" as="xs:string"/>
      <xsl:element name="validity">
         <xsl:attribute name="validFromDate" select="$validFromDate"/>
         <xsl:attribute name="validToDate" select="$validToDate"/>
      </xsl:element>
   </xsl:function>

   <xsl:template name="shared:write">
      <xsl:param name="output" as="map(*)"/>
      <xsl:for-each select="map:keys($output)">
         <xsl:sort select="."/>
         <xsl:sequence select="shared:write(., map:get($output, .))"/>
      </xsl:for-each>
   </xsl:template>

   <xsl:function name="shared:write">
      <xsl:param name="href" as="xs:string"/>
      <xsl:param name="value" as="xs:string"/>
      <!-- Exclude default added "output" when base-output-uri is absent by checking is-absolute-uri() -->
      <xsl:if test="shared:is-absolute-uri($href)">
         <xsl:sequence select="shared:log('DEBUG', concat('Exporting ', $href))"/>
         <xsl:sequence select="file:write-text($href, $value)"/>
      </xsl:if>
   </xsl:function>

</xsl:stylesheet>